import logging
import multiprocessing
from typing import Iterator, NamedTuple, Optional

from more_itertools import partition

from .error import IDLParseError
from .webidl import parse, validate
from .webidl.nodes import Ast


logger = logging.getLogger(__name__)


class ParseResult(NamedTuple):

    item: Optional[Ast] = None
    error: Optional[str] = None


def parse_idl(file: str) -> ParseResult:
    try:
        return ParseResult(item=parse(file))
    except IDLParseError:
        return ParseResult(error='Skipped {file}\n{errors}'.format(
            file=file, errors='\n'.join(map(str, validate(file)))))
    except Exception as exc:
        return ParseResult(error=str(exc))


def process_idl(idl_files: tuple[str, ...]) -> list[Ast]:
    return IDLProcessor(idl_files).process()


class IDLProcessor:

    def __init__(self, idl_files: tuple[str, ...], process: int = 4):
        self._idl_files = idl_files
        self._process = process

    def process(self) -> list[Ast]:
        return self.parse()

    def _parse(self) -> Iterator[ParseResult]:
        with multiprocessing.Pool(self._process) as pool:
            yield from pool.imap_unordered(parse_idl, self._idl_files)

    def parse(self) -> list[Ast]:
        successes, fails = partition(
            pred=lambda result: result.item is None, iterable=self._parse())

        for fail in fails:
            logger.debug(fail.error)

        return [success.item for success in successes]  # type: ignore



import copy
from functools import reduce
from typing import Dict, TypeVar


_T = TypeVar('_T')
GRAPH = Dict[_T, set[_T]]


def get_unknown_dependencies(graph: GRAPH[_T]) -> set[_T]:
    return reduce(set.union, graph.values(), set()) - set(graph)  # type: ignore


def topological_order(graph: GRAPH[_T]) -> tuple[tuple[set[_T], ...], GRAPH[_T]]:
    """
    TODO: return `raw_graph` split into cycle and unreachable.
    """
    _ = get_unknown_dependencies(graph)

    raw_graph = copy.deepcopy(graph)
    for item, dependencies in raw_graph.items():
        dependencies.discard(item)

    resolved = []
    while True:
        if not (ordered := set(item for item, deps in raw_graph.items() if not deps)):
            break

        resolved.append(ordered)

        raw_graph = {
            item: deps - ordered
            for item, deps in raw_graph.items()
            if item not in ordered
        }

    return tuple(resolved), raw_graph



from typing import final

from more_itertools.more import first

from .nodes import (
    Argument,
    Attribute,
    Callback,
    CallbackInterface,
    Const,
    Constructor,
    Definitions,
    Dictionary,
    Enum,
    ExtendedAttribute,
    Field,
    IdlType,
    Includes,
    Infinity,
    Interface,
    InterfaceMixin,
    Iterable_,
    Literal,
    LiteralList,
    MapLike,
    Namespace,
    NaN,
    Null,
    Operation,
    SetLike,
    Typedef,
    Value,
)
from .generated import WebIDLParser, WebIDLParserVisitor
from .utils import setup_type


@final
class Visitor(WebIDLParserVisitor):  # pylint: disable=too-many-public-methods

    def __init__(self, tree):
        self._tree = tree

    def run(self):
        return self.visit(self._tree)

    def visitWebIDL(self, ctx: WebIDLParser.WebIDLContext):
        return ctx.definitions().accept(self)

    def visitDefinitions(self, ctx: WebIDLParser.DefinitionsContext):
        return Definitions(
            definitions=[
                definition.accept(self)
                for definition in ctx.extendedDefinition()
            ],
        )

    def visitExtendedDefinition(self, ctx: WebIDLParser.ExtendedDefinitionContext):
        definition = ctx.definition().accept(self)

        if extended_attribute := ctx.extendedAttributeList():
            definition.ext_attrs = extended_attribute.accept(self)

        return definition

    def visitDictionary(self, ctx: WebIDLParser.DictionaryContext):
        return Dictionary(
            partial=ctx.PARTIAL() is not None,
            name=ctx.IDENTIFIER().getText(),
            members=[member.accept(self) for member in ctx.dictionaryMembers()],
            inheritance=ctx.inheritance() and ctx.inheritance().accept(self),
        )

    def visitTypedef(self, ctx: WebIDLParser.TypedefContext):
        setup_type(
            idl_type := ctx.typeWithExtendedAttributes().accept(self),
            'typedef-type',
        )

        return Typedef(idl_type=idl_type, name=ctx.IDENTIFIER().getText())

    def visitIncludesStatement(self, ctx: WebIDLParser.IncludesStatementContext):
        return Includes(
            target=ctx.target.text, includes=ctx.includes.text)  # type: ignore

    def visitEnum_(self, ctx: WebIDLParser.Enum_Context):
        return Enum(
            name=ctx.IDENTIFIER().getText(),
            values=[
                Literal(
                    type='enum-value', value=enum_value.getText().strip('"'))
                for enum_value in ctx.StringLiteral()
            ]
        )

    def visitInterfaceRest(self, ctx: WebIDLParser.InterfaceRestContext):
        return Interface(
            name=ctx.IDENTIFIER().getText(),
            inheritance=ctx.inheritance() and ctx.inheritance().accept(self),
            members=[member.accept(self) for member in ctx.interfaceMembers()],
        )

    def visitPartialInterfaceRest(self, ctx: WebIDLParser.PartialInterfaceRestContext):
        return Interface(
            name=ctx.IDENTIFIER().getText(),
            members=[
                member.accept(self) for member in ctx.partialInterfaceMembers()
            ],
        )

    def visitMixinRest(self, ctx: WebIDLParser.MixinRestContext):
        return InterfaceMixin(
            name=ctx.IDENTIFIER().getText(),
            members=[
                member.accept(self) for member in ctx.mixinMembers()
            ],
        )

    def visitNamespace(self, ctx: WebIDLParser.NamespaceContext):
        return Namespace(
            name=ctx.IDENTIFIER().getText(),
            members=[
                member.accept(self) for member in ctx.namespaceMembers()
            ],
            partial=ctx.PARTIAL() is not None,
        )

    def visitPartial(self, ctx: WebIDLParser.PartialContext):
        interface_or_mixin = ctx.partialInterfaceOrPartialMixin().accept(self)

        interface_or_mixin.partial = True

        return interface_or_mixin

    def _member_with_extended_attribute(self, member, extended_attribute):
        member = member.accept(self)

        if extended_attribute is not None:
            member.ext_attrs = extended_attribute.accept(self)

        return member

    def visitNamespaceMembers(self, ctx: WebIDLParser.NamespaceMembersContext):
        return self._member_with_extended_attribute(
            ctx.namespaceMember(), ctx.extendedAttributeList())

    def visitMixinMembers(self, ctx: WebIDLParser.MixinMembersContext):
        return self._member_with_extended_attribute(
            ctx.mixinMember(), ctx.extendedAttributeList())

    def visitPartialInterfaceMembers(self, ctx: WebIDLParser.PartialInterfaceMembersContext):
        return self._member_with_extended_attribute(
            ctx.partialInterfaceMember(), ctx.extendedAttributeList())

    def visitInterfaceMembers(self, ctx: WebIDLParser.InterfaceMembersContext):
        return self._member_with_extended_attribute(
            ctx.interfaceMember(), ctx.extendedAttributeList())

    def visitDictionaryMembers(self, ctx: WebIDLParser.DictionaryMembersContext):
        return self._member_with_extended_attribute(
            ctx.dictionaryMember(), ctx.extendedAttributeList())

    def visitCallbackInterfaceMembers(self, ctx: WebIDLParser.CallbackInterfaceMembersContext):
        return self._member_with_extended_attribute(
            ctx.callbackInterfaceMember(), ctx.extendedAttributeList())

    def visitMixinMember(self, ctx: WebIDLParser.MixinMemberContext):
        if ctx.READONLY() is None:
            return ctx.getChild(0).accept(self)

        attribute = ctx.attributeRest().accept(self)
        attribute.readonly = True

        return attribute

    def visitDictionaryMember(self, ctx: WebIDLParser.DictionaryMemberContext):
        if type_ := ctx.typeWithExtendedAttributes():
            idl_type = type_.accept(self)
        else:
            idl_type = ctx.type_().accept(self)

        if default := ctx.default_():
            default = default.accept(self)

        setup_type(idl_type, 'dictionary-type')

        return Field(
            required=ctx.REQUIRED() is not None,
            name=ctx.IDENTIFIER().getText(),
            default=default,
            idl_type=idl_type,
        )

    def visitNamespaceMember(self, ctx: WebIDLParser.NamespaceMemberContext):
        if regular_operation := ctx.regularOperation():
            return self._operation(regular_operation)

        attribute = ctx.attributeRest().accept(self)
        attribute.readonly = True

        return attribute

    def visitStringifier(self, ctx: WebIDLParser.StringifierContext):
        if stringifier := ctx.stringifierRest():
            stringifier = stringifier.accept(self)
            stringifier.special = 'stringifier'

            return stringifier

        return Operation(name='', idl_type=None, special='stringifier')

    def visitStringifierRest(self, ctx: WebIDLParser.StringifierRestContext):
        if regular_operation := ctx.regularOperation():
            return self._operation(regular_operation)

        attribute = ctx.attributeRest().accept(self)
        attribute.readonly = ctx.READONLY() is not None

        return attribute

    def visitStaticMember(self, ctx: WebIDLParser.StaticMemberContext):
        member = ctx.staticMemberRest().accept(self)
        member.special = 'static'

        return member

    def visitStaticMemberRest(self, ctx: WebIDLParser.StaticMemberRestContext):
        if regular_operation := ctx.regularOperation():
            return self._operation(regular_operation)

        attribute = ctx.attributeRest().accept(self)
        attribute.readonly = ctx.READONLY() is not None

        return attribute

    def visitCallbackOrInterfaceOrMixin(self, ctx: WebIDLParser.CallbackOrInterfaceOrMixinContext):
        if callback := ctx.callbackRestOrInterface():
            return callback.accept(self)

        return ctx.interfaceOrMixin().accept(self)

    def visitCallbackRestOrInterface(self, ctx: WebIDLParser.CallbackRestOrInterfaceContext):
        if callback_rest := ctx.callbackRest():
            return callback_rest.accept(self)

        return CallbackInterface(
            name=ctx.IDENTIFIER().getText(),
            inheritance=None,
            members=[
                member.accept(self)
                for member in ctx.callbackInterfaceMembers()
            ],
        )

    def visitInterfaceOrMixin(self, ctx: WebIDLParser.InterfaceOrMixinContext):
        if interface_rest := ctx.interfaceRest():
            return interface_rest.accept(self)

        return ctx.mixinRest().accept(self)

    def _operation(self, regular_operation, special=''):
        return_type, (name, arguments) = regular_operation.accept(self)

        setup_type(return_type, 'return-type')

        return Operation(
            name=name,
            arguments=arguments,
            idl_type=return_type,
            special=special,
        )

    def visitMaplikeRest(self, ctx: WebIDLParser.MaplikeRestContext):
        return MapLike(
            idl_type=[
                type_with_extended_attr.accept(self)
                for type_with_extended_attr in ctx.typeWithExtendedAttributes()
            ],
            readonly=ctx.READONLY() is not None,
        )

    def visitSetlikeRest(self, ctx: WebIDLParser.SetlikeRestContext):
        return SetLike(
            idl_type=[ctx.typeWithExtendedAttributes().accept(self)],
            readonly=ctx.READONLY() is not None,
        )

    def visitCallbackInterfaceMember(self, ctx: WebIDLParser.CallbackInterfaceMemberContext):
        if regular_operation := ctx.regularOperation():
            return self._operation(regular_operation)

        return ctx.const_()

    def visitConstValue(self, ctx: WebIDLParser.ConstValueContext):
        if boolean_literal := ctx.booleanLiteral():
            return boolean_literal.accept(self)

        if float_literal := ctx.floatLiteral():
            return float_literal.accept(self)

        return Literal(type='number', value=ctx.IntegerLiteral().getText())

    def visitFloatLiteral(self, ctx: WebIDLParser.FloatLiteralContext):
        if ctx.NAN():
            return NaN()

        if ctx.INFINITY():
            return Infinity(negative=False)

        if ctx.MINUS_INFINITY():
            return Infinity(negative=True)

        return Literal(type='number', value=ctx.DecimalLiteral().getText())

    def visitBooleanLiteral(self, ctx: WebIDLParser.BooleanLiteralContext):
        return Literal(type='boolean', value=ctx.getText() == 'true')

    def visitConst_(self, ctx: WebIDLParser.Const_Context):
        return Const(
            idl_type=ctx.constType().accept(self),
            value=ctx.constValue().accept(self),
            name=ctx.IDENTIFIER().getText(),
        )

    def visitCallbackRest(self, ctx: WebIDLParser.CallbackRestContext):
        if arguments := ctx.argumentList():
            arguments = arguments.accept(self)

        setup_type(idl_type := ctx.returnType().accept(self), 'return-type')

        return Callback(
            name=ctx.IDENTIFIER().getText(),
            idl_type=idl_type,
            arguments=arguments or [],
        )

    def visitConstructor(self, ctx: WebIDLParser.ConstructorContext):
        if arguments := ctx.argumentList():
            arguments = arguments.accept(self)

        return Constructor(arguments=arguments or [])

    def visitAttributeRest(self, ctx: WebIDLParser.AttributeRestContext):
        setup_type(
            idl_type := ctx.typeWithExtendedAttributes().accept(self),
            'attribute-type',
        )

        return Attribute(
            name=ctx.attributeName().accept(self), idl_type=idl_type)

    def visitIterable(self, ctx: WebIDLParser.IterableContext):
        if arg_list := ctx.optionalArgumentList():
            arg_list = arg_list.accept(self)

        return Iterable_(
            arguments=arg_list or [],
            async_=ctx.ASYNC() is not None,
            idl_type=[
                type_with_extended_attr.accept(self)
                for type_with_extended_attr in ctx.typeWithExtendedAttributes()
            ],
        )

    def visitOptionalArgumentList(self, ctx: WebIDLParser.OptionalArgumentListContext):
        if arg_list := ctx.argumentList():
            arg_list = arg_list.accept(self)

        return arg_list or []

    def visitOperation(self, ctx: WebIDLParser.OperationContext):
        return self._operation(
            ctx.regularOperation(), ctx.special and ctx.special.text or '')

    def visitRegularOperation(self, ctx: WebIDLParser.RegularOperationContext):
        return ctx.returnType().accept(self), ctx.operationRest().accept(self)

    def visitOperationRest(self, ctx: WebIDLParser.OperationRestContext):
        if name := ctx.operationName():
            name = name.accept(self)

        if arguments := ctx.argumentList():
            arguments = arguments.accept(self)

        return name or '', arguments or []

    def visitArgumentList(self, ctx: WebIDLParser.ArgumentListContext):
        return [argument.accept(self) for argument in ctx.argument()]

    def visitArgument(self, ctx: WebIDLParser.ArgumentContext):
        argument = ctx.argumentRest().accept(self)

        if ext_attrs := ctx.extendedAttributeList():
            argument.ext_attrs = ext_attrs.accept(self)

        return argument

    def visitExtendedAttributeList(self, ctx: WebIDLParser.ExtendedAttributeListContext):
        return [
            extendedAttribute.accept(self)
            for extendedAttribute in ctx.extendedAttribute()
        ]

    @staticmethod
    def _extended_attribute(name, rhs=None, arguments=None):
        return ExtendedAttribute(name=name, rhs=rhs, arguments=arguments or [])

    def visitExtendedAttributeNoArgs(self, ctx: WebIDLParser.ExtendedAttributeNoArgsContext):
        return self._extended_attribute(ctx.IDENTIFIER().getText())

    def visitExtendedAttributeNamedArgList(
            self, ctx: WebIDLParser.ExtendedAttributeNamedArgListContext):
        return self._extended_attribute(
            name=ctx.name.text,  # type: ignore
            arguments=ctx.argumentList().accept(self),
            rhs=Literal(type='identifier', value=ctx.rhs.text),  # type: ignore
        )

    def visitExtendedAttributeArgList(self, ctx: WebIDLParser.ExtendedAttributeArgListContext):
        return self._extended_attribute(
            name=ctx.name.text,  # type: ignore
            arguments=ctx.argumentList().accept(self),
        )

    def visitExtendedAttributeIdentList(self, ctx: WebIDLParser.ExtendedAttributeIdentListContext):
        return self._extended_attribute(
            name=ctx.name.text,  # type: ignore
            rhs=ctx.identifierList().accept(self)
        )

    def visitExtendedAttributeIdent(self, ctx: WebIDLParser.ExtendedAttributeIdentContext):
        return self._extended_attribute(
            name=ctx.name.text, rhs=ctx.rhs.accept(self))  # type: ignore

    def visitIdentifierList(self, ctx: WebIDLParser.IdentifierListContext):
        identifiers = [
            identifier.accept(self) for identifier in ctx.identifier()
        ]

        return LiteralList(
            value=[Value(value=identifier.value) for identifier in identifiers],
            type=f'{first(identifiers).type}-list',
        )

    def visitArgumentRest(self, ctx: WebIDLParser.ArgumentRestContext):
        if default := ctx.default_():
            default = default.accept(self)

        if optional := ctx.OPTIONAL():
            idl_type = ctx.typeWithExtendedAttributes().accept(self)
        else:
            idl_type = ctx.type_().accept(self)

        setup_type(idl_type, 'argument-type')

        return Argument(
            idl_type=idl_type,
            name=ctx.argumentName().accept(self),
            optional=optional is not None,
            default=default,
            variadic=ctx.ELLIPSIS() is not None,
        )

    def visitReadOnlyMember(self, ctx: WebIDLParser.ReadOnlyMemberContext):
        member = ctx.readOnlyMemberRest().accept(self)
        member.readonly = True

        return member

    def visitInheritAttribute(self, ctx: WebIDLParser.InheritAttributeContext):
        attribute = ctx.attributeRest().accept(self)
        attribute.special = 'inherit'

        return attribute

    def visitTypeWithExtendedAttributes(
            self, ctx: WebIDLParser.TypeWithExtendedAttributesContext):
        idl_type = ctx.type_().accept(self)

        if ext_attrs := ctx.extendedAttributeList():
            idl_type.ext_attrs = ext_attrs.accept(self)

        return idl_type

    def visitDefault_(self, ctx: WebIDLParser.Default_Context):
        return ctx.defaultValue().accept(self)

    def visitDefaultValue(self, ctx: WebIDLParser.DefaultValueContext):
        if str_ := ctx.StringLiteral():
            return Literal(type='string', value=str_.getText().strip('"'))

        if const_value := ctx.constValue():
            return const_value.accept(self)

        if ctx.LEFT_BRACE():
            return Literal(type='dictionary', value=[])

        if ctx.NULL():
            return Null()

        return Literal(type='sequence', value=[])

    def visitType_(self, ctx: WebIDLParser.Type_Context):
        if single_type := ctx.singleType():
            return single_type.accept(self)

        return IdlType(
            idl_type=ctx.unionType().accept(self),
            nullable=ctx.null_() is not None,
            union=True,
        )

    def visitUnionType(self, ctx: WebIDLParser.UnionTypeContext):
        return [
            member_type.accept(self) for member_type in ctx.unionMemberType()
        ]

    def visitUnionMemberType(self, ctx: WebIDLParser.UnionMemberTypeContext):
        nullable = ctx.null_() is not None

        if ext_attrs := ctx.extendedAttributeList():
            ext_attrs = ext_attrs.accept(self)

        if distinguishable_type := ctx.distinguishableType():
            return IdlType(
                idl_type=distinguishable_type.accept(self),
                ext_attrs=ext_attrs or [],
                nullable=nullable,
            )

        if generic := ctx.genericType():
            return IdlType(
                idl_type=[generic.accept(self)],
                ext_attrs=ext_attrs or [],
                nullable=nullable
            )

        return IdlType(
            idl_type=ctx.unionType().accept(self),
            nullable=nullable,
            union=True,
        )

    def visitReturnType(self, ctx: WebIDLParser.ReturnTypeContext):
        if void := ctx.VOID():
            return IdlType(idl_type=void.getText())

        return ctx.type_().accept(self)

    def visitUnsignedIntegerType(
            self, ctx: WebIDLParser.UnsignedIntegerTypeContext):
        integer_type = ctx.integerType().accept(self)

        if ctx.UNSIGNED():
            integer_type = f'unsigned {integer_type}'

        return integer_type

    def visitSingleType(self, ctx: WebIDLParser.SingleTypeContext):
        nullable = ctx.null_() is not None

        if any_ := ctx.ANY():
            return IdlType(idl_type=any_.getText())

        if promise := ctx.promiseType():
            return IdlType(idl_type=[promise.accept(self)], generic='Promise')

        if generic := ctx.genericType():
            return IdlType(
                idl_type=[generic.accept(self)],
                generic='sequence',
                nullable=nullable,
            )

        distinguishable_type = ctx.distinguishableType().accept(self)

        if isinstance(distinguishable_type, IdlType):
            return distinguishable_type

        return IdlType(
            idl_type=distinguishable_type, nullable=nullable)

    def visitGenericType(self, ctx: WebIDLParser.GenericTypeContext):
        return ctx.typeWithExtendedAttributes().accept(self)

    def visitDistinguishableType(
            self, ctx: WebIDLParser.DistinguishableTypeContext):
        if type_ := ctx.primitiveType():
            return type_.accept(self)

        if type_ := ctx.stringType():
            return type_.accept(self)

        if type_ := ctx.bufferRelatedType():
            return type_.accept(self)

        if type_ := ctx.recordType():
            return type_.accept(self)

        return ctx.getChild(0).getText()

    def visitRecordType(self, ctx: WebIDLParser.RecordTypeContext):
        return IdlType(
            generic='record',
            idl_type=[
                IdlType(idl_type=ctx.stringType().accept(self)),
                ctx.typeWithExtendedAttributes().accept(self),
            ]
        )

    def visitConstType(self, ctx: WebIDLParser.ConstTypeContext):
        if idl_type := ctx.primitiveType():
            idl_type = idl_type.accept(self)

        return IdlType(
            type='const-type', idl_type=idl_type or ctx.IDENTIFIER().getText())

    def visitPrimitiveType(self, ctx: WebIDLParser.PrimitiveTypeContext):
        if type_ := ctx.unsignedIntegerType():
            return type_.accept(self)

        if type_ := ctx.unrestrictedFloatType():
            return type_.accept(self)

        return ctx.getChild(0).getText()

    def visitUnrestrictedFloatType(self, ctx: WebIDLParser.UnrestrictedFloatTypeContext):
        float_type = ctx.floatType().accept(self)

        if ctx.UNRESTRICTED():
            float_type = f'unrestricted {float_type}'

        return float_type

    def visitPromiseType(self, ctx: WebIDLParser.PromiseTypeContext):
        return ctx.returnType().accept(self)

    def visitIntegerType(self, ctx: WebIDLParser.IntegerTypeContext):
        if long := ' '.join(long.getText() for long in ctx.LONG()):
            return long

        return ctx.getText()

    def visitFloatType(self, ctx: WebIDLParser.FloatTypeContext):
        return ctx.getText()

    def visitStringType(self, ctx: WebIDLParser.StringTypeContext):
        return ctx.getText()

    def visitOperationName(self, ctx: WebIDLParser.OperationNameContext):
        return ctx.getText()

    def visitArgumentName(self, ctx: WebIDLParser.ArgumentNameContext):
        return ctx.getText()

    def visitAttributeName(self, ctx: WebIDLParser.AttributeNameContext):
        return ctx.getText()

    def visitInheritance(self, ctx: WebIDLParser.InheritanceContext):
        return ctx.IDENTIFIER().getText()

    def visitOther(self, ctx: WebIDLParser.OtherContext):
        if ctx.IntegerLiteral():
            type_ = 'integer'
        elif ctx.DecimalLiteral():
            type_ = 'decimal'
        elif ctx.StringLiteral():
            type_ = 'string'
        else:
            type_ = 'identifier'

        return Literal(type=type_, value=ctx.getText())



from .const import CATCH_CONSTANT, LET
from .nodes import (
    AssignmentExpression,
    BlockStatement,
    Expression,
    ExpressionStatement,
    CallExpression,
    CatchClause,
    Identifier,
    Literal,
    MemberExpression,
    NewExpression,
    TryStatement,
    VariableDeclaration,
    VariableDeclarator,
    ObjectExpression,
    Property,
)


def variable_ast(name: str, expression: Expression) -> VariableDeclaration:
    return VariableDeclaration(
        kind=LET,
        declarations=[
            VariableDeclarator(id=Identifier(name=name), init=expression),
        ]
    )


def create_expression(name: str, expression: Expression) -> ExpressionStatement:
    return ExpressionStatement(
            expression=AssignmentExpression(
                left=Identifier(name=name),
                right=expression,
            )
    )


def create_object(name, progenitor, arguments) -> ExpressionStatement:
    return create_expression(
        name=name,
        expression=NewExpression(
            callee=Identifier(name=progenitor),
            arguments=arguments,
        )
    )


def create_dict(name, properties) -> ExpressionStatement:
    return create_expression(
        name=name,
        expression=ObjectExpression(
            properties=properties,
        ),
    )


def create_attribute(name, progenitor, method) -> ExpressionStatement:
    return create_expression(
        name=name,
        expression=MemberExpression(
            object=Identifier(name=progenitor),
            property=Identifier(name=method),
        )
    )


def create_operation(name, progenitor, method, arguments) -> ExpressionStatement:
    return create_expression(
        name=name,
        expression=CallExpression(
            arguments=arguments,
            callee=MemberExpression(
                object=Identifier(name=progenitor),
                property=Identifier(name=method),
            ),
        )
    )


def try_statement(var: VariableDeclaration) -> TryStatement:
    return TryStatement(
        block=BlockStatement(body=[var]),
        handler=CatchClause(
            param=Identifier(name=CATCH_CONSTANT),
            block=BlockStatement(body=[]),
        ),
    )


def create_literal(value) -> Literal:
    return Literal(value=value)


def create_identifier(name) -> Identifier:
    return Identifier(name=name)


def create_property(key, value) -> Property:
    return Property(key=key, value=value)



from contextlib import contextmanager
from types import SimpleNamespace
from typing import Dict, Iterator, NewType

from .js.nodes import Ast as JsAst
from .visitor import Visitor
from .utils import interleave


Parenthesis = NewType('Parenthesis', str)

PAREN = Parenthesis('paren')
BRACE = Parenthesis('brace')
BRACKET = Parenthesis('bracket')

_parentheses: Dict[Parenthesis, SimpleNamespace] = {
    PAREN: SimpleNamespace(open='(', close=')'),
    BRACE: SimpleNamespace(open='{', close='}'),
    BRACKET: SimpleNamespace(open='[', close=']'),
}


class Unparser(Visitor[JsAst]):
    """
    Methods in this class recursively traverse an AST and output source code
    for the abstract syntax.
    """
    def __init__(self):
        self._source = []

    def write(self, text):
        self._source.append(text)

    def traverse(self, node):
        if isinstance(node, list):
            for item in node:
                self.traverse(item)
        else:
            super().visit(node)

    def visit(self, node) -> str:
        self.traverse(node)

        return ''.join(self._source)

    def visit_identifier(self, node):
        self.write(node.name)

    def visit_variable_declaration(self, node):
        self.write(f'{node.kind} ')

        self.traverse(node.declarations)

    def visit_variable_declarator(self, node):
        self.traverse(node.id)

        self.write(' = ')

        self.traverse(node.init)

    def visit_new_expression(self, node):
        self.write('new ')
        self.traverse(node.callee)

        with self._parentheses(PAREN):
            interleave(
                iterable=node.arguments,
                func=self.traverse,
                separator=lambda: self.write(', ')
            )

    def visit_object_expression(self, node):
        with self._parentheses(BRACE):
            interleave(
                iterable=node.properties,
                func=self.traverse,
                separator=lambda: self.write(', ')
            )

    def visit_property(self, node):
        self.write(node.key)

        self.write(': ')

        self.traverse(node.value)

    def visit_member_expression(self, node):
        self.traverse(node.object)

        self.write('.')

        self.traverse(node.property)

    def visit_call_expression(self, node):
        self.traverse(node.callee)

        with self._parentheses(PAREN):
            interleave(
                iterable=node.arguments,
                func=self.traverse,
                separator=lambda: self.write(', ')
            )

    def visit_literal(self, node):
        self.write(node.raw)

    def visit_try_statement(self, node):
        self.write('try ')

        with self._parentheses(BRACE):
            self.traverse(node.block)

        self.traverse(node.handler)

    def visit_catch_clause(self, node):
        self.write(' catch')

        with self._parentheses(PAREN):
            self.traverse(node.param)

        with self._parentheses(BRACE):
            self.traverse(node.block)

    def visit_block_statement(self, node):
        interleave(
            iterable=node.body,
            func=self.traverse,
            separator=lambda: self.write('\n')
        )

    def visit_assignment_expression(self, node):
        self.traverse(node.left)

        self.write(' = ')

        self.traverse(node.right)

    @contextmanager
    def _parentheses(self, paren: Parenthesis) -> Iterator[None]:
        self.write(_parentheses[paren].open)
        yield
        self.write(_parentheses[paren].close)


def unparse(ast: JsAst) -> str:
    return Unparser().visit(ast)



from pathlib import Path

from idl2js.storage import Storage
from idl2js.unparser import unparse
from idl2js.idl_processor import process_idl
from idl2js.js.statements import try_statement
from idl2js.built_in_types import BuiltInTypes
from idl2js.builder import Builder, DefinitionStorage, build


class Idl2Js:

    def __init__(self, idl: tuple[str, ...]):
        self._storage = Storage()
        self._std_types = BuiltInTypes()
        self._prepared_idl = process_idl(idl)
        self._definition_storage = DefinitionStorage(self._prepared_idl)

        self._builder = Builder(
            std_types=self._std_types,
            definition_storage=self._definition_storage,
        )

    def generate(self):
        return [
            unparse(try_statement(variable.ast))
            for variable in build(self._definition_storage, self._builder)
        ]


def main():
    # pylint: disable=import-outside-toplevel
    from pprint import pprint
    raw_idl = (Path(__file__).parent.parent / 'webidl' / 'blob.webidl').resolve()
    raw_idl2 = (Path(__file__).parent.parent / 'webidl' / 'BlobEvent.webidl').resolve()
    idl2js = Idl2Js(idl=(str(raw_idl), str(raw_idl2)))
    pprint(idl2js.generate(), width=220)


if __name__ == '__main__':
    main()



import string
from random import randint, choice


BYTE = 'byte'
OCTET = 'octet'
SHORT = 'short'
UNSIGNED_SHORT = 'unsigned short'
LONG = 'long'
UNSIGNED_LONG = 'unsigned long'
LONG_LONG = 'long long'
UNSIGNED_LONG_LONG = 'unsigned long long'

DOM_STRING = 'DOMString'
USV_STRING = 'USVString'

INT = frozenset((
    BYTE,
    OCTET,
    SHORT,
    UNSIGNED_SHORT,
    LONG,
    UNSIGNED_LONG,
    LONG_LONG,
    UNSIGNED_LONG_LONG,
))

STRING = frozenset((
    DOM_STRING,
    USV_STRING,
))


_INT_RANGES = {
    BYTE: [-2 ** 7, 2 ** 7 - 1],  # int8
    OCTET: [0, 2 ** 8 - 1],  # uint8
    SHORT: [-2 ** 15, 2 ** 15 - 1],  # int16
    UNSIGNED_SHORT: [0, 2 ** 16 - 1],  # uint16
    LONG: [-2 ** 31, 2 ** 31 - 1],  # int32
    UNSIGNED_LONG: [0, 2 ** 32 - 1],  # uint32
    LONG_LONG: [-2 ** 63, 2 ** 63 - 1],  # int64
    UNSIGNED_LONG_LONG: [0, 2 ** 64 - 1],  # uint64
}


class BuiltInTypes:

    def __contains__(self, item):
        return item in INT | STRING

    def generate(self, type_):
        if type_ in INT:
            return self.generate_int(type_)

        if type_ == DOM_STRING:
            return self.generate_string()

        return type_

    @staticmethod
    def generate_int(int_type):
        return randint(*_INT_RANGES[int_type])

    @staticmethod
    def generate_string():
        return ''.join(
            choice(string.ascii_lowercase)
            for _ in range(randint(1, 10))
        )



from typing import Any, Dict, List

import attr

from .nodes import Ast
from .parser import Parser, SyntaxErrorInfo
from .visitor import Visitor


def validate(file: str) -> List[SyntaxErrorInfo]:
    return Parser(file).validate()


def parse(file: str) -> Ast:
    return Visitor(Parser(file).parse()).run()


def parse_as_dict(file: str) -> Dict[str, Any]:
    return attr.asdict(parse(file))



import abc
from typing import Any, List, Optional, Union

import attr


@attr.s
class Ast(abc.ABC):
    ...


@attr.s
class Module(Ast):

    type: str = attr.ib(default='module')
    source_type: str = attr.ib(default='Program')
    body: List[Any] = attr.ib(factory=list)


@attr.s
class NewExpression(Ast):

    callee = attr.ib()
    arguments = attr.ib(factory=list)
    type = attr.ib(default='NewExpression')


@attr.s
class VariableDeclaration(Ast):

    type: str = attr.ib(default='VariableDeclaration')
    kind: str = attr.ib(default='let')
    declarations: List[Any] = attr.ib(factory=list)


@attr.s
class VariableDeclarator(Ast):

    id = attr.ib()
    init = attr.ib()
    type: str = attr.ib(default='VariableDeclarator')


@attr.s
class Identifier(Ast):

    name: str = attr.ib()
    type: str = attr.ib(default='Identifier')


@attr.s
class MemberExpression(Ast):

    object = attr.ib()
    property = attr.ib()
    computed: bool = attr.ib(default=False)
    type: str = attr.ib(default='MemberExpression')


@attr.s
class CallExpression(Ast):

    callee = attr.ib()
    arguments = attr.ib(factory=list)
    type: str = attr.ib(default='CallExpression')


@attr.s
class Literal(Ast):

    raw = attr.ib(init=False)
    value = attr.ib()
    type: str = attr.ib(default='Literal')

    def __attrs_post_init__(self):
        if isinstance(self.value, int):
            self.raw = str(self.value)
        elif isinstance(self.value, str):
            self.raw = f"'{self.value}'"
        else:
            self.raw = self.value


@attr.s
class BlockStatement(Ast):

    body: List[Any] = attr.ib(factory=list)
    type: str = attr.ib(default='BlockStatement')


@attr.s
class CatchClause:
    param: Identifier = attr.ib()
    block: BlockStatement = attr.ib()
    type: str = attr.ib(default='CatchClause')


@attr.s
class TryStatement(Ast):

    handler = attr.ib()
    block: BlockStatement = attr.ib()
    finalizer: Optional[Any] = attr.ib(default=None)
    type: str = attr.ib(default='TryStatement')


@attr.s
class AssignmentExpression(Ast):

    left: Identifier = attr.ib()
    right: Any = attr.ib()
    operator: str = attr.ib(default='=')
    type: str = attr.ib(default='AssignmentExpression')


@attr.s
class ExpressionStatement(Ast):

    expression: Any = attr.ib()
    type: str = attr.ib(default='ExpressionStatement')


@attr.s
class ArrayExpression(Ast):
    ...


@attr.s
class Property(Ast):

    key: Any = attr.ib()
    value: Any = attr.ib()
    type: str = attr.ib(default='Property')
    method: bool = attr.ib(default=False)
    shorthand: bool = attr.ib(default=False)
    computed: bool = attr.ib(default=False)


@attr.s
class ObjectExpression(Ast):

    type: str = attr.ib(default='ObjectExpression')
    properties: List[Property] = attr.ib(factory=list)


Expression = Union[
    AssignmentExpression,
    NewExpression,
    MemberExpression,
    CallExpression,
    ObjectExpression,
]
